from .response import *
from .order import *

class OrdersResponse(Response):
    def __init__(self, account_id, response_format, data):
        super().__init__(response_format, data)
        self.account_id = account_id
        self.orders = []
        if response_format.lower() == 'xml':
            self.__parse_xml(data)
        elif response_format.lower() == 'json':
            self.__parse_json(data)

    def __parse_xml(self, data):
        super().parse_xml(data)
        for order_xml in self.xml.find('orderstatus').findall('order'):
            order = Order()
            order.from_xml(order_xml)
            self.orders.append(order)

    def __parse_json(self, data):
        super().parse_json(data)
        for order_json in self.json['orderstatus']['order']:
            order = Order()
            order.from_json(order_json)
            self.orders.append(order)

    def get_orders(self):
        return self.orders




class Holding():
    def __init__(self):
        pass

    def from_xml(self, xml):
        pass

    def from_json(self, json):
        # Feeling thankful for Notepad++ macros...
        if 'accounttype' in json:
        	self.accounttype = json['accounttype']
        if 'assetclass' in json:
        	self.assetclass = json['assetclass']
        if 'cfi' in json:
        	self.cfi = json['cfi']
        if 'change' in json['quote']:
        	self.change = json['quote']['change']
        if 'costbasis' in json:
        	self.costbasis = json['costbasis']
        if 'cusip' in json['instrument']:
        	self.cusip = json['instrument']['cusip']
        if 'desc' in json['instrument']:
        	self.desc = json['instrument']['desc']
        if 'factor' in json['instrument']:
        	self.factor = json['instrument']['factor']
        if 'gainloss' in json:
        	self.gainloss = json['gainloss']
        if 'lastprice' in json['quote']:
        	self.lastprice = json['quote']['lastprice']
        if 'marketvalue' in json:
        	self.marketvalue = json['marketvalue']
        if 'marketvaluechange' in json:
        	self.marketvaluechange = json['marketvaluechange']
        if 'matdt' in json:
        	self.matdt = json['matdt']
        if 'mmy' in json:
        	self.mmy = json['mmy']
        if 'mult' in json:
        	self.mult = json['mult']
        if 'price' in json:
        	self.price = json['price']
        if 'purchaseprice' in json:
        	self.purchaseprice = json['purchaseprice']
        if 'putcall' in json['instrument']:
        	self.putcall = json['instrument']['putcall']
        if 'qty' in json:
        	self.qty = json['qty']
        if 'sectyp' in json['instrument']:
        	self.sectyp = json['instrument']['sectyp']
        if 'strkpx' in json:
        	self.strkpx = json['strkpx']
        if 'sym' in json['instrument']:
        	self.sym = json['instrument']['sym']
        if 'totalsecurities' in json:
        	self.totalsecurities = json['totalsecurities']



from .request import *
from ..responses.quotes import *

class QuotesRequest(Request):
    def __init__(self, symbols=[], fids=[], response_format='json'):
        super().__init__(response_format)
        self.symbols = symbols
        self.fids = fids

    def set_fids(self, fids):
        self.fids = fids

    def set_symbols(self, symbols):
        self.symbols = symbols

    def get_symbols(self):
        return self.symbols

    def get_fids(self):
        return self.fids

    def execute(self, allyApi):
        return QuotesResponse(self.response_format, allyApi.get_quote(self.symbols))



from xml.etree import ElementTree

def get_class_vars(class_type):
    return [v for k, v in vars(class_type).items() if \
            not k.startswith('_')]

class Order():
    def __init__(self, **data):
        # From Request Object.
        self.acct = data.get('acct')
        # Stock Symbol.
        self.sym = data.get('sym')
        # Number of Shares.
        self.qty = str(int(data.get('qty')))
        # use class SECURITY_TYPE
        self.sec_typ = str(data.get('sec_typ', ''))
        # use class SIDE
        self.side = str(data.get('side', ''))
        # use class ORDER_TYPE
        self.typ = str(data.get('typ', ''))

        # For Changes and Cancels
        self.ord_id = data.get('ord_id')

        # For Buying to Cover a Short Position.
        self.acct_typ = data.get('acct_typ')

        # For Limit Orders
        # Limit Price.
        self.px = f"{data.get('px', 0.0):.2f}"
        # use class TIME_IN_FORCE
        self.tm_in_force = str(data.get('tm_in_force', ''))

        # For Options
        # use class OPTION_POSITION
        self.pos_efct = str(data.get('pos_efct', ''))
        # Strike Price.
        self.strk_px = f"{data.get('strk_px', 0.0):.2f}"
        # use class OPTION_CLASS
        self.cfi = str(data.get('cfi', ''))
        # Date of Maturity.
        self.mat_dt = data.get('mat_dt')
        # Option Expiration.
        self.mmy = data.get('mmy')

    def validate(self):
        """Verify all required information is in the order.
        """
        # Account must exist.
        assert self.acct
        # Symbol must exist.
        assert self.sym
        # Quantity must exist and be an integer greater than zero.
        # Partials are sold when # shares held is less than one.
        assert int(self.qty) > 0
        # Order type must exist and be in ORDER_TYPE.
        assert self.typ
        assert self.typ in get_class_vars(ORDER_TYPE)
        # Side must exist and be in SIDE.
        assert self.side
        assert self.side in get_class_vars(SIDE)
        # Security type must exist and be in SECURITY_TYPE.
        assert self.sec_typ
        assert self.sec_typ in get_class_vars(SECURITY_TYPE)

        # If Account Type is used, it must be in ACCOUNT_TYPE.
        if self.acct_typ:
            assert self.acct_type in get_class_vars(ACCOUNT_TYPE)

        if self.typ != ORDER_TYPE.MARKET:
            # Time in Force must exist and be in TIME_IN_FORCE.
            assert self.tm_in_force
            assert self.tm_in_force in get_class_vars(TIME_IN_FORCE)
        if self.typ in [ORDER_TYPE.LIMIT, ORDER_TYPE.STOP_LIMIT]:
            # Price must exist and be a float greater than zero.
            assert self.px
            assert float(self.px) > 0.0

        if self.sec_typ == SECURITY_TYPE.OPTION:
            # Position must exist and be in OPTION_POSITION.
            assert self.pos_efct
            assert self.pos_efct in get_class_vars(OPTION_POSITION)
            # Strike Price must exist & be a float greater than zero.
            assert self.strk_px
            assert float(self.strk_px) > 0.0
            # CFI must exist and be in OPTION_CLASS.
            assert self.cfi
            assert self.cfi in get_class_vars(OPTION_CLASS)
            # Date of Maturity must exist.
            assert self.mat_dt
            # Option Expiration must exist.
            assert self.mmy

        return True

    def from_xml(self, xml):
        nsp = {'': 'http://www.fixprotocol.org/FIXML-5-0-SP2'}
        exec_rpt = xml.find('FIXML', nsp).find('ExecRpt', nsp)
        if not exec_rpt:
          return
        self.ord_id = exec_rpt.attrib.get('OrdID')
        self.id = exec_rpt.attrib.get('ID')
        self.stat = exec_rpt.attrib.get('Stat')
        self.acct = exec_rpt.attrib.get('Acct')
        self.acct_typ = exec_rpt.attrib.get('AcctTyp')
        self.side = exec_rpt.attrib.get('Side')
        self.typ = exec_rpt.attrib.get('Typ')
        self.px = exec_rpt.attrib.get('Px')
        self.tm_in_force = exec_rpt.attrib.get('TmInForce')
        self.leaves_qty = exec_rpt.attrib.get('LeavesQty')
        self.trd_dt = exec_rpt.attrib.get('TrdDt')
        self.txn_tm = exec_rpt.attrib.get('TxnTm')
        instrmt = exec_rpt.find('Instrmt', nsp)
        if instrmt:
            self.sym = instrmt.attrib.get('Sym')
            self.sec_typ = instrmt.attrib.get('SecTyp')
            self.desc = instrmt.attrib.get('Desc')
        ordqty = exec_rpt.find('OrdQty', nsp)
        if ordqty:
            self.qty = ordqty.attrib.get('Qty')
        comm = exec_rpt.find('Comm', nsp)
        if comm:
            self.comm = comm.attrib.get('Comm')

    def from_json(self, json):
        exec_rpt = json.get('FIXML', {}).get('ExecRpt')
        if not exec_rpt:
          return
        self.ord_id = exec_rpt.get('@OrdID')
        self.id = exec_rpt.get('@ID')
        self.stat = exec_rpt.get('@Stat')
        self.acct = exec_rpt.get('@Acct')
        self.acct_typ = exec_rpt.get('@AcctTyp')
        self.side = exec_rpt.get('@Side')
        self.typ = exec_rpt.get('@Typ')
        self.px = exec_rpt.get('@Px')
        self.tm_in_force = exec_rpt.get('@TmInForce')
        self.leaves_qty = exec_rpt.get('@LeavesQty')
        self.trd_dt = exec_rpt.get('@TrdDt')
        self.txn_tm = exec_rpt.get('@TxnTm')
        instrmt = exec_rpt.get('Instrmt')
        if instrmt:
            self.sym = instrmt.get('@Sym')
            self.sec_typ = instrmt.get('@SecTyp')
            self.desc = instrmt.get('@Desc')
        ordqty = exec_rpt.get('OrdQty')
        if ordqty:
            self.qty = ordqty.get('@Qty')
        comm = exec_rpt.get('Comm')
        if comm:
            self.comm = comm.get('@Comm')

    def to_fixml(self, cancel=False):
        """Convert the contents of the order to FIXML.
           This is only for common stock and single-leg options.
            @param self - the object pointer
            @param cancel - should this order be cancelled only?
        """
        nsp = {'xmlns': 'http://www.fixprotocol.org/FIXML-5-0-SP2'}
        order_tag = "Order"
        if self.ord_id and cancel:
            order_tag = "OrdCxlReq"
        elif self.ord_id:
            order_tag = "OrdCxlRplcReq"

        # Now Build the FIXML.
        # First get the attributes always required.
        base_xml = ElementTree.Element("FIXML", nsp)
        order = ElementTree.SubElement(base_xml, order_tag)
        order.set('Acct', self.acct)
        order.set('Typ', self.typ)
        order.set('Side', self.side)
        instrmt = ElementTree.SubElement(order, 'Instrmt')
        instrmt.set('SecTyp', self.sec_typ)
        instrmt.set('Sym', self.sym)
        ordqty = ElementTree.SubElement(order, 'OrdQty')
        ordqty.set('Qty', self.qty)

        # Add field-dependent attributes.
        if self.ord_id:
            order.set('OrigID', self.ord_id)
        if self.side == SIDE.BUY and \
           self.acct_typ == ACCOUNT_TYPE.SHORT:
            order.set('AcctTyp', self.acct_typ)
        if self.typ != ORDER_TYPE.MARKET:
            order.set('TmInForce', self.tm_in_force)
            if self.typ != ORDER_TYPE.STOP:
                order.set('Px', self.px)
        if self.sec_typ == SECURITY_TYPE.OPTION:
            order.set('PosEfct', self.pos_efct)
            instrmt.set('CFI', self.cfi)
            instrmt.set('StrkPx', self.strk_px)
            instrmt.set('MMY', self.mmy)
            instrmt.set('MatDt', self.mat_dt.isoformat())

        return base_xml

def get_multileg_fixml(orders, cancel=False):
    """Create FIXML for a multi-leg option chain using multiple orders.
        @param orders - A list of the orders.
        @param cancel - should this order be cancelled only?
    """
    if not orders:
        return None
    # Validate all the orders have the same basic info.
    orders[0].validate()
    chk_dict = dict(ord_id=orders[0].ord_id,
                    typ=orders[0].typ,
                    tm_in_force=orders[0].tm_in_force,
                    px=orders[0].px,
                    acct=orders[0].acct,
                    sym=orders[0].sym,
                   )
    for order in orders[1:]:
        # Along with validation, all basic data must match.
        order.validate()
        assert order.ord_id == chk_dict['ord_id']
        assert order.typ == chk_dict['typ']
        assert order.tm_in_force == chk_dict['tm_in_force']
        assert order.px == chk_dict['px']
        assert order.acct == chk_dict['acct']
        assert order.sym == chk_dict['sym']

    # Set the Namespace and base tag name.
    nsp = {'xmlns': 'http://www.fixprotocol.org/FIXML-5-0-SP2'}
    order_tag = "NewOrdMLeg"
    if chk_dict['ord_id'] and cancel:
        order_tag = "OrdCxlReq"
    elif chk_dict['ord_id']:
        order_tag = "MLegOrdCxlRplc"

    # Now Build the FIXML.
    # First get the attributes always required.
    base_xml = ElementTree.Element("FIXML", nsp)
    mleg = ElementTree.SubElement(base_xml, order_tag)
    mleg.set('Acct', chk_dict['acct'])
    if chk_dict['ord_id']:
        # For a replace or cancel, the original ID is needed.
        mleg.set('OrigCIOrdID', chk_dict['ord_id'])
    if chk_dict['ord_id'] and cancel:
        # For a cancel, the FIXML is much simpler.
        instrmt = ElementTree.SubElement(mleg, 'Instrmt')
        instrmt.set('SecTyp', SECURITY_TYPE.MLEG)
        instrmt.set('Sym', chk_dict['sym'])
    else:
        # For all others, fill in the rest of the info for each leg.
        mleg.set('OrdTyp', chk_dict['typ'])
        if chk_dict['typ'] == ORDER_TYPE.LIMIT:
            mleg.set('TmInForce', chk_dict['tm_in_force'])
            mleg.set('Px', chk_dict['px'])
        for order in orders:
            # Cycle through each order and add it.
            ord_el = ElementTree.SubElement(mleg, 'Ord')
            ord_el.set('OrdQty', order.qty)
            ord_el.set('PosEfct', order.pos_efct)
            leg = ElementTree.SubElement(ord_el, 'Leg')
            leg.set('Side', order.side)
            leg.set('Strk', order.strk_px)
            leg.set('Mat', order.mat_dt.isoformat())
            leg.set('MMY', order.mmy)
            leg.set('SecTyp', order.sec_typ)
            leg.set('CFI', order.cfi)
            leg.set('Sym', order.sym)

    return base_xml

class TIME_IN_FORCE:
    DAY = "0"
    GTC = "1"
    MARKET_ON_CLOSE = "7"

class ACCOUNT_TYPE:
    SHORT = "5"

class ORDER_TYPE:
    MARKET = "1"
    LIMIT = "2"
    STOP = "3"
    STOP_LIMIT = "4"

class SIDE:
    BUY = "1"
    SELL = "2"
    SELL_SHORT = "5"

class SECURITY_TYPE:
    COMMON_STOCK = "CS"
    OPTION = "OPT"
    MULTI_LEG = "MLEG"

class OPTION_POSITION:
    OPEN = "O"
    CLOSE = "C"

class OPTION_CLASS:
    CALL = "OC"
    PUT = "OP"



"""@package AllyAPI
    A Python3 class that allows access to all of the functionality in the
    Ally/TradeKing API.

    This package attempts to stay on top of changes to the API and allow an
    easy to user iterface with the Ally Invest API. The API does no formatting
    for the user. A response format of 'xml' or 'json' can be specified and
    the API responses will be returned as the raw XML or JSON, respectively.

    This API was built with the developer in mind and should allow a developer
    to build applications around the Ally Invest API without having to deal with
    accessing and managing the requests and responses.

    This project was inspired my PyAlly (https://github.com/alienbrett/PyAlly).
"""

from requests_oauthlib import OAuth1
from xml.etree import ElementTree
import datetime
import requests
import json
from collections import defaultdict
import re
import copy

from .URLs import URLs

class AllyAPI:
    """The AllyAPI class providing blackbox use of the Ally Invest API.

    This is the main class of the API module. This should be the only class used in
    applictions built around the API. The AllyAPI class allows access to the GET and
    POST requests supported by Ally Invest.

    Missing Functionality:
        MARKET
            GET market/options/search
            GET market/options/strikes
            GET market/options/expirations
            GET market/timesales
        WATCHLIST
            GET watchlists/:id
            DELETE watchlists/:id
            POST watchlists/:id/symbols
            DELETE watchlists/:id/symbols
        STREAMING OPERATIONS
            MARKET
                GET market/quotes
    """
    def __init__(self, oauth_secret, oauth_token, client_key,
                response_format="json"):
        """AllyAPI constructor. Sets the response format on all of the URLs and
            the oauth/client keys required to access the API.

            Parameters
                @param self - the object pointer
                @param oauth_secret - secret oauth key from Ally
                @param oauth_token - oauth token from Ally
                @param client_key - client key from Ally
                @param response_format - format of the response. Valid values are 'xml' and 'json'.
                    Specifying 'xml' will return an ElementTree containing the response XML while
                    'json' will return the response in the JSON format.
        """
        self.format = response_format
        self.url = URLs(response_format=response_format)

        self.oauth_secret = oauth_secret
        self.oauth_token = oauth_token
        self.client_key = client_key
        self.client_secret = client_key

        self.auth_time = None
        self.auth = None
        self.valid_auth_dt = datetime.timedelta(seconds=10)

    def __create_auth(self):
        """A private method to create the OAuth1 object, if necessary."""
        now = datetime.datetime.now()

        if self.auth == None or self.auth_time + self.valid_auth_dt < now:
            self.auth_time = now
            self.auth = OAuth1(self.client_key, self.client_secret, self.oauth_token,
                          self.oauth_secret, signature_type='auth_header')

    def __get_symbol_string(self, symbols):
        """Returns a string that is either a single quote or a comma-separated
            list of quotes depending on the type of quotes.
            @param self - the object pointer
            @param symbols - single ticker or list of ticker to get quotes for
        """
        if not isinstance(symbols, str): # list
            symbols = ",".join(symbols)
        return symbols

    def __convert_fixml_json(self, json_data):
        """Takes the order data and converts it to a consistent format.
           The FIXML message is also expanded, with the original intact.
            @param self - the object pointer
            @param json_data - original data to be converted.
        """
        # If there's no orders, there's nothing to do.
        if not json_data["response"]["orderstatus"]["order"]:
            return json_data

        # Copy the data to keep from overwriting the input.
        data = copy.deepcopy(json_data)

        # A single order will be a dict, and multiple a list.
        # Convert order to always be a list of dicts.
        if isinstance(data["response"]["orderstatus"]["order"], dict):
            data["response"]["orderstatus"]["order"] = \
                [data["response"]["orderstatus"]["order"]]

        # Convert the FIXML message in each order.
        # Add the keys to order itself, but preserve fixmlmessage.
        for order in data["response"]["orderstatus"]["order"]:
            order_xml = ElementTree.fromstring(order["fixmlmessage"])
            order.update(self.__fixml_to_dict(order_xml))

        # Return the converted data.
        return data

    def __convert_fixml_xml(self, xml_data):
        """Takes the order data and expands the FIXML message.
           The original message is left intact.
            @param self - the object pointer
            @param xml_data - original data to be converted.
        """
        # Register the FIXML namespace.
        ElementTree.register_namespace("", "http://www.fixprotocol.org/FIXML-5-0-SP2")

        # Copy the data to keep from overwriting the input.
        data = copy.deepcopy(xml_data)

        # Each order will have a "fixmlmessage" to convert.
        for order in data.find("orderstatus").findall("order"):
            fixml_text = order.find("fixmlmessage").text
            fixml = ElementTree.fromstring(fixml_text)
            order.append(fixml)

        # Return the converted data.
        return data

    def __fixml_to_dict(self, fixml):
        """Recursively convert FIXML to a dictionary.
            @param self - the object pointer
            @param fixml - FIXML Element to be converted.
        """
        # Remove the Namespace from the tag.
        tag = re.sub(r"\{[^}]*\} *", "", fixml.tag)

        # Establish the final dictionary to return.
        ret = {tag: {} if fixml.attrib else None}

        # Recursively convert each subelement.
        # Each subelement becomes a tag key with a dict value.
        children = list(fixml)
        if children:
            defdict = defaultdict(list)
            for childdict in map(self.__fixml_to_dict, children):
                for key, val in childdict.items():
                    defdict[key].append(val)
            ret = {tag: {k: v[0] if len(v) == 1 else v
                           for k, v in defdict.items()}}

        # Set each attribute as a tag key.
        if fixml.attrib:
            ret[tag].update(("@" + k, v)
                              for k, v in fixml.attrib.items())

        # Set the value of each attribute key to the text.
        if fixml.text:
            text = fixml.text.strip()
            if children or fixml.attrib:
                if text:
                  ret[tag]["#text"] = text
            else:
                ret[tag] = text

        # Return the final dictionary.
        return ret

    def __get_fixml(self, ticker, amount, type, account, side, tif, price, sectype):
        fixml = "<FIXML xmlns=\"http://www.fixprotocol.org/FIXML-5-0-SP2\">"
        fixml += "<Order"
        if type != ORDER_TYPE.MARKET and tif is not None:
            fixml += " TmInForce=\"{}\"".format(tif)
        if type != ORDER_TYPE.MARKET:
            fixml += " Px=\"{}\"".format(price)
        fixml += " Typ=\"{}\" Side=\"{}\" Acct=\"{}\">".format(type, side, account)
        fixml += "<Instrmt SecTyp=\"{}\" Sym=\"{}\"/>".format(sectype, ticker)
        fixml += "<OrdQty Qty=\"{}\"/></Order></FIXML>".format(amount)

        return fixml

    def __to_format(self, response, xml=False):
        """A private method to return the API response in the desired format
            @param self - the object pointer
            @param response - response from the Ally Invest API
        """
        if response.status_code != 200:
            if response.status_code == 429:
                print("Too many requests.")
                exit()
            elif response.status_code == 414:
                print("URI too long, please chunk ticker symbols.")
                exit()
        if self.format == "json" and not xml:
            return response.json()
        else:
            return ElementTree.fromstring(response.content)

    def __get_data(self, url):
        """A private method to return the requested data in the requested format
            for a given URL.
            @param self - the object pointer
            @param url - API URL to access
        """
        self.__create_auth()
        return self.__to_format(requests.get(url, auth=self.auth))

    def __submit_post(self, url, data, headers={}, usexml=False):
        """A private method to submit a post request to the Ally Invest server
            @param self - the object pointer
            @param url - API URL to access
            @param data - payload for the HTTP request
        """
        self.__create_auth()
        res = requests.post(url, headers=headers, data=data, auth=self.auth)
        return self.__to_format(res, usexml)

    def get_accounts(self):
        """Returns all of the user's accounts."""
        return self.__get_data(self.url.accounts_url())

    def get_accounts_balances(self):
        """Returns the balances of all of the user's accounts."""
        return self.__get_data(self.url.accounts_balances_url())

    def get_account(self, id):
        """Returns a specific account provided the account ID (account number)
            @param self - the object pointer
            @param id - account number
        """
        return self.__get_data(self.url.account_url().format(id=str(id)))

    def get_account_balances(self, id):
        """Returns the balances of a specific account (ID = account number)
            @param self - the object pointer
            @param id - account number
        """
        return self.__get_data(self.url.account_balances_url().format(id=str(id)))

    def get_account_history(self, id):
        """Returns the history of a specific account (ID = account number)
            @param self - the object pointer
            @param id - account number
        """
        return self.__get_data(self.url.account_history_url().format(id=str(id)))

    def get_account_holdings(self, id):
        """Returns the holdings of a specific account (ID = account number)
            @param self - the object pointer
            @param id - account number
        """
        return self.__get_data(self.url.account_holdings_url().format(id=str(id)))

    def get_orders(self, id):
        """Returns the orders of a specific account (ID = account number)
            @param self - the object pointer
            @param id - account number
        """
        data = self.__get_data(self.url.get_orders().format(id=str(id)))
        if self.format == "json":
          return self.__convert_fixml_json(data)

        return self.__convert_fixml_xml(data)

    def post_order(self, id, fixml):
        """Posts an order and returns the response.
            @param self - the object pointer
            @param id - account number
            @param fixml - FIXML string to send.
        """
        headers = {
            'TKI_OVERRIDE': 'true',
            'Content-Type': 'application/xml',
        }
        # The GET and POST have the same URL.
        url = self.url.get_orders().format(id=str(id))
        return self.__submit_post(url, fixml, headers,
                                  self.format=='xml')

    def post_order_preview(self, id, fixml):
        """Posts an order for preview and returns the response.
            @param self - the object pointer
            @param id - account number
            @param fixml - FIXML string to send.
        """
        headers = {
            'TKI_OVERRIDE': 'true',
            'Content-Type': 'application/xml',
        }
        url = self.url.post_order_preview().format(id=str(id))
        return self.__submit_post(url, fixml, headers,
                                  self.format=='xml')

    def get_market_clock(self):
        """Returns the state of the market, the time until next state change,
            and current server timestamp.
            @param self - the object pointer
        """
        return self.__get_data(self.url.clock_url())

    def get_quote(self, symbols):
        """Returns quote information for a single ticker or list of tickers.
            Note: this function does not implement selecting customer FIDs as
            described in the API documentation. These can be filtered from the return
            if need be.
            @param self - the object pointer
            @param symbols - single ticker or list of ticker to get quotes for
        """
        url = self.url.quote_url()+"?symbols={symbols}"
        symbols = self.__get_symbol_string(symbols)
        return self.__get_data(url.format(symbols=symbols))

    def __get_option_quote_symbol(self, symbol, exp_date, strike, put_call):
        sym = "{sym}{year}{month:02d}{day:02d}{putcall}{strike}"
        strike = str(int(strike*1000)).zfill(8)
        return sym.format(sym=symbol.upper(), year=str(exp_date.year)[-2:], month=exp_date.month,
                        day=exp_date.day, putcall=put_call.upper(), strike=strike)

    def get_option_quote(self, symbol, expiration_date, strike_price, put_call):
        """Returns a quote for an option for the symbol, expiration date, strike price
            and put/call specifier.

            @param self - object pointer
            @param symbol - underlying stock's ticker symbol
            @param expiration_date - options expiration date
            @param strike_price - option's strike price
            @param put_call - c=call, p=put
        """
        url = self.url.quote_url() + "?symbols={sym}"

        if isinstance(symbol, str): # single ticker
            if not isinstance(expiration_date, datetime.datetime):
                print("In 'get_option_quote': datetime.datetime expected for expiration date.")
                return None
            sym = self.__get_option_quote_symbol(symbol, expiration_date, strike_price, put_call)
        elif isinstance(symbol, list) and isinstance(expiration_date, list) \
            and isinstance(strike_price, list) and isinstance(put_call, list):
            if not isinstance(expiration_date[0], datetime.datetime):
                print("In 'get_option_quote': datetime.datetime expected for expiration date.")
                return None
            request_sym = []
            for i in range(len(symbol)):
                request_sym.append(self.__get_option_quote_symbol(symbol[i], expiration_date[i],
                            strike_price[i], put_call[i]))
            sym = self.__get_symbol_string(request_sym)
        else:
            print("In 'get_option_quote': symbol, expiration_date, strike_price, and put_call \
                  must all be single values or lists.")
            return None

        return self.__get_data(url.format(sym=sym))

    def news_search(self, symbols, startdate=None, enddate=None, maxhits=10):
        """Retrieves a listing of news headlines based on symbols.
            @param self - the object pointer
            @param symbols - single ticker or list of ticker to get quotes for
            @param startdate - search for articles between this date and enddate
            @param enddate - search for articles between this date and startdate
            @param maxhits - number of articles to return
        """
        if startdate is None or enddate is None:
            print("news_search: either enddate or startdate is not specified, ignoring both.")

        if (startdate is not None and enddate is not None) and (enddate < startdate):
            print("news_search: start date is after end date.")
            raise Exception("Start date is after end date in news search.")

        url = self.url.news_search_url() + "?symbols={syms}&maxhits={mxhits}".format(mxhits=maxhits, syms="{syms}")
        if startdate is not None and enddate is not None:
            url += "&startdate={sdate}&enddate={edate}" \
                .format(sdate=startdate.strftime("%m/%d/%Y"), edate=enddate.strftime("%m/%d/%Y"))

        return self.__get_data(url.format(syms=self.__get_symbol_string(symbols)))

    def get_news_article(self, article_id):
        """Gets a single news article based on the article ID. This ID can be retrieved
            from the news_search()  function.
            @param self - the object pointer
            @param article_id - ID of the article to retrieve
        """
        return self.__get_data(self.url.news_article_url().format(article_id=article_id))

    def get_toplists(self, listtype="topgainers", exchange="N"):
        """Returns a ranked list depending on listtype and exchange.
            @param listtype - type of list to be queried, accepted values are:
                'toplosers': top losers by dollar amount
                'toppctlosers': top percentage losers
                'topvolume': top volume
                'topactive': top active
                'topgainers': top gainers by dollar amount (default)
                'toppctgainers': top percentage gainers
            @param exchange - exchange to be queried, accepted values are:
                'A': American Stock Exchange
                'N': New York Stock Exchange (default)
                'Q': NASDAQ
                'U': NASDAQ Bulletin Board
                'V': NASDAQ OTC Other
        """
        url = self.url.toplists_url().format(listtype=listtype)
        url += "?exchange={ex}".format(ex=exchange)
        return self.__get_data(url)

    def get_options(self, symbol):
        url = self.url.options_search_url() + ("?symbol={}".format(symbol))
        return self.__get_data(url)

    def get_options_strikes(self, symbol):
        url = self.url.options_strikes_url() + ("?symbol={}".format(symbol))
        return self.__get_data(url)

    def get_options_expirations(self, symbol):
        url = self.url.options_exps_url() + ("?symbol={}".format(symbol))
        return self.__get_data(url)

    def get_member_profile(self):
        """Returns general information associated with the user including account
            numbers and account information.
            @param self - the object pointer
        """
        return self.__get_data(self.url.member_profile_url())

    def get_status(self):
        """Returns an error if the API endpoint/server is unavailable. Otherwise
            returns the current server timestamp.
            @param self - the object pointer
        """
        return self.__get_data(self.url.status_url())

    def get_version(self):
        """Gets the current version of the API of the endpoint called.
            @param self - the object pointer
        """
        return self.__get_data(self.url.version_url())

    def get_watchlists(self):
        """Retrieves all watchlists belonging to the member.
            @param self - the object pointer
        """
        return self.__get_data(self.url.get_watchlists_url())


    def create_watchlist(self, watchlist_name, symbols=""):
        """Creates a watchlist and adds a symbol or list of symbols to a watchlist.
            WARNING: There appears to be an issue when adding a list of symbols.
                It is recommended that one ticker symbol is added at a time.
            @param self - the object pointer
            @param watchist_id - name of the watchlist
            @param symbols - single ticker or list of tickers to add to the watchlist
        """
        print("WARNING create_watchlist(): There appears to be an issue when adding a list of symbols. It is recommended that one ticker symbol is added at a time.")
        payload = {"id": watchlist_name}
        if not symbols == "":
            payload["symbols"] = symbols
        return self.__submit_post(self.url.post_watchlist_url(), payload)

    def order_common_stock(self, ticker, shares, type, account_nbr, side,
                            time_in_force=None, price=None):
        """Creates an order for common stock (as opposed to options).
            @param self - object pointer
            @param ticker - ticker symbol of the security to purchase
            @param shares - the number of shares to purchase
            @param type - the order type: Market, Limit, Stop, or Stop Limit
                - use the provided enum for these values
            @param account_nbr - the account number for which the shares are to be purchased
            @param side - the side of the trade: Buy or Sell
                - use the provided enum for these values
            @param time_in_force - not applicable for market orders: Day Order, Good til Cancelled, Market on Close
                - use the provided enum for these values
            @param price - the price to purchase the security (only for limit and stop limit orders)
        """
        if price == None and type != ORDER_TYPE.MARKET:
            raise("Price is required for non-market order types.")
        payload = self.__get_fixml(ticker, shares, type, account_nbr, side, time_in_force, price, "CS")
        headers = {
            'TKI_OVERRIDE': 'true',
            'Content-Type': 'application/xml',
        }
        url = self.url.get_post_order().format(id=account_nbr)
        return self.__submit_post(url, payload, headers, True)

class TIME_IN_FORCE:
    DAY = "0"
    GTC = "1"
    MARKET_ON_CLOSE = "7"

class ORDER_TYPE:
    MARKET = "1"
    LIMIT = "2"
    STOP = "3"
    STOP_LIMIT = "4"

class SIDE:
    BUY = "1"
    SELL = "2"
    SELL_SHORT = "5"



import json

class Response():
    def __init__(self, response_format, data):
        self.response_format = response_format
        self.raw_data = data
        self.error = ''

    def parse_xml(self, data):
        self.xml = data
        self.error = data.find('error').text

    def parse_json(self, data):
        self.json = json.loads(json.dumps(data["response"]))
        self.error = self.json['error']

    def get_raw_data(self):
        return self.raw_data



from xml.etree import ElementTree

from .request import *
from ..responses.post_order_preview import *
from ..responses.order import *

class PostOrderPreviewRequest(Request):
    def __init__(self, account_id, order, response_format='json'):
        super().__init__(response_format)
        self.account_id = account_id
        self.order = order

    def execute(self, ally_api, cancel=False):
        if isinstance(self.order, list):
            # This is a multi-leg option order.
            # Never trust order contents; always validate it.
            for order in self.order:
                order.validate()
            # Convert to string.
            fixml_string = ElementTree.tostring(get_multileg_fixml( self.order, cancel=cancel))
        else:
            # This is a single-leg option or common stock order.
            # Never trust order contents; always validate it.
            self.order.validate()
            # Convert to string.
            fixml_string = ElementTree.tostring(self.order.to_fixml(cancel=cancel))
        return PostOrderPreviewResponse(self.account_id, self.response_format, ally_api.post_order_preview(self.account_id, fixml_string))



from .request import *
from ..responses.quotes import *

class OptionQuoteRequest(Request):
    def __init__(self, symbol, exp_date, strike, put_call, response_format='json'):
        super().__init__(response_format)
        self.symbol = symbol
        self.exp_date = exp_date
        self.strike = strike
        self.put_call = put_call

    def execute(self, allyApi):
        return QuotesResponse(self.response_format, allyApi.get_option_quote(self.symbol, self.exp_date, self.strike, self.put_call))



from .response import *

class PostOrderPreviewResponse(Response):
    def __init__(self, account_id, response_format, data):
        super().__init__(response_format, data)
        self.account_id = account_id
        if response_format.lower() == 'xml':
            self.__parse_xml(data)
        elif response_format.lower() == 'json':
            self.__parse_json(data)

    def __parse_xml(self, data):
        super().parse_xml(data)

    def __parse_json(self, data):
        super().parse_json(data)



class Quote():
    def __init__(self):
        pass

    def from_xml(self, xml):
        pass

    def from_json(self, json):
        if 'adp_100' in json:
            self.adp_100 = json['adp_100']
        if 'adp_200' in json:
            self.adp_200 = json['adp_200']
        if 'adp_50' in json:
            self.adp_50 = json['adp_50']
        if 'adv_21' in json:
            self.adv_21 = json['adv_21']
        if 'adv_30' in json:
            self.adv_30 = json['adv_30']
        if 'adv_90' in json:
            self.adv_90 = json['adv_90']
        if 'ask' in json:
            self.ask = json['ask']
        if 'ask_time' in json:
            self.ask_time = json['ask_time']
        if 'asksz' in json:
            self.asksz = json['asksz']
        if 'basis' in json:
            self.basis = json['basis']
        if 'beta' in json:
            self.beta = json['beta']
        if 'bid' in json:
            self.bid = json['bid']
        if 'bid_time' in json:
            self.bid_time = json['bid_time']
        if 'bidsz' in json:
            self.bidsz = json['bidsz']
        if 'bidtick' in json:
            self.bidtick = json['bidtick']
        if 'chg' in json:
            self.chg = json['chg']
        if 'chg_sign' in json:
            self.chg_sign = json['chg_sign']
        if 'chg_t' in json:
            self.chg_t = json['chg_t']
        if 'cl' in json:
            self.cl = json['cl']
        if 'contract_size' in json:
            self.contract_size = json['contract_size']
        if 'cusip' in json:
            self.cusip = json['cusip']
        if 'date' in json:
            self.date = json['date']
        if 'datetime' in json:
            self.datetime = json['datetime']
        if 'days_to_expiration' in json:
            self.days_to_expiration = json['days_to_expiration']
        if 'div' in json:
            self.div = json['div']
        if 'divexdate' in json:
            self.divexdate = json['divexdate']
        if 'divfreq' in json:
            self.divfreq = json['divfreq']
        if 'divpaydt' in json:
            self.divpaydt = json['divpaydt']
        if 'dollar_value' in json:
            self.dollar_value = json['dollar_value']
        if 'eps' in json:
            self.eps = json['eps']
        if 'exch' in json:
            self.exch = json['exch']
        if 'exch_desc' in json:
            self.exch_desc = json['exch_desc']
        if 'hi' in json:
            self.hi = json['hi']
        if 'iad' in json:
            self.iad = json['iad']
        if 'idelta' in json:
            self.idelta = json['idelta']
        if 'igamma' in json:
            self.igamma = json['igamma']
        if 'imp_volatility' in json:
            self.imp_volatility = json['imp_volatility']
        if 'incr_vl' in json:
            self.incr_vl = json['incr_vl']
        if 'irho' in json:
            self.irho = json['irho']
        if 'issue_desc' in json:
            self.issue_desc = json['issue_desc']
        if 'itheta' in json:
            self.itheta = json['itheta']
        if 'ivega' in json:
            self.ivega = json['ivega']
        if 'last' in json:
            self.last = json['last']
        if 'lo' in json:
            self.lo = json['lo']
        if 'name' in json:
            self.name = json['name']
        if 'op_delivery' in json:
            self.op_delivery = json['op_delivery']
        if 'op_flag' in json:
            self.op_flag = json['op_flag']
        if 'op_style' in json:
            self.op_style = json['op_style']
        if 'op_subclass' in json:
            self.op_subclass = json['op_subclass']
        if 'openinterest' in json:
            self.openinterest = json['openinterest']
        if 'opn' in json:
            self.opn = json['opn']
        if 'opt_val' in json:
            self.opt_val = json['opt_val']
        if 'pchg' in json:
            self.pchg = json['pchg']
        if 'pchg_sign' in json:
            self.pchg_sign = json['pchg_sign']
        if 'pcls' in json:
            self.pcls = json['pcls']
        if 'pe' in json:
            self.pe = json['pe']
        if 'phi' in json:
            self.phi = json['phi']
        if 'plo' in json:
            self.plo = json['plo']
        if 'popn' in json:
            self.popn = json['popn']
        if 'pr_adp_100' in json:
            self.pr_adp_100 = json['pr_adp_100']
        if 'pr_adp_200' in json:
            self.pr_adp_200 = json['pr_adp_200']
        if 'pr_adp_50' in json:
            self.pr_adp_50 = json['pr_adp_50']
        if 'pr_date' in json:
            self.pr_date = json['pr_date']
        if 'pr_openinterest' in json:
            self.pr_openinterest = json['pr_openinterest']
        if 'prbook' in json:
            self.prbook = json['prbook']
        if 'prchg' in json:
            self.prchg = json['prchg']
        if 'prem_mult' in json:
            self.prem_mult = json['prem_mult']
        if 'put_call' in json:
            self.put_call = json['put_call']
        if 'pvol' in json:
            self.pvol = json['pvol']
        if 'qcond' in json:
            self.qcond = json['qcond']
        if 'rootsymbol' in json:
            self.rootsymbol = json['rootsymbol']
        if 'secclass' in json:
            self.secclass = json['secclass']
        if 'sesn' in json:
            self.sesn = json['sesn']
        if 'sho' in json:
            self.sho = json['sho']
        if 'strikeprice' in json:
            self.strikeprice = json['strikeprice']
        if 'symbol' in json:
            self.symbol = json['symbol']
        if 'tcond' in json:
            self.tcond = json['tcond']
        if 'timestamp' in json:
            self.timestamp = json['timestamp']
        if 'tr_num' in json:
            self.tr_num = json['tr_num']
        if 'tradetick' in json:
            self.tradetick = json['tradetick']
        if 'trend' in json:
            self.trend = json['trend']
        if 'under_cusip' in json:
            self.under_cusip = json['under_cusip']
        if 'undersymbol' in json:
            self.undersymbol = json['undersymbol']
        if 'vl' in json:
            self.vl = json['vl']
        if 'volatility12' in json:
            self.volatility12 = json['volatility12']
        if 'vwap' in json:
            self.vwap = json['vwap']
        if 'wk52hi' in json:
            self.wk52hi = json['wk52hi']
        if 'wk52hidate' in json:
            self.wk52hidate = json['wk52hidate']
        if 'wk52lo' in json:
            self.wk52lo = json['wk52lo']
        if 'wk52lodate' in json:
            self.wk52lodate = json['wk52lodate']
        if 'xdate' in json:
            self.xdate = json['xdate']
        if 'xday' in json:
            self.xday = json['xday']
        if 'xmonth' in json:
            self.xmonth = json['xmonth']
        if 'xyear' in json:
            self.xyear = json['xyear']
        if 'yield' in json:
            self.yld = json['yield']



from .response import *

class AccountsBalancesResponse(Response):
    def __init__(self, response_format, data):
        super().__init__(response_format, data)
        self.account_data = {}      # a dictionary of dictionaries
        if response_format.lower() == 'xml':
            self.__parse_xml(data)
        elif response_format.lower() == 'json':
            self.__parse_json(data)

    def get_account_data_by_account(self, id):
        return self.account_data[str(id)]

    def get_all_account_data(self):
        return self.account_data

    def __parse_xml(self, data):
        pass

    def __parse_json(self, data):
        super().parse_json(data)
        for account in self.json["accountbalance"]:
            this_data = {}
            this_data["accountname"] = account["accountname"]
            this_data["accountvalue"] = account["accountvalue"]
            self.account_data[account["account"]] = this_data
